/* $Id: name6.c,v 1.1.2.3.2.5 1998/07/09 05:03:50 onoe Exp $ */
/*
 *	Atsushi Onoe <onoe@sm.sony.co.jp>
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>

#include <netdb.h>
#include <resolv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>

#ifndef _PATH_HOSTS
#define	_PATH_HOSTS	"/etc/hosts"
#endif

#ifndef MAXALIASES
#define	MAXALIASES	10
#endif
#ifndef	MAXADDRS
#define	MAXADDRS	20
#endif
#ifndef MAXDNAME
#define	MAXDNAME	1025
#endif

#define	ADDRLEN(af)	((af) == AF_INET6 ? sizeof(struct in6_addr) : \
					    sizeof(struct in_addr))
#define	MAPADDR(ab, ina) {						\
	memcpy(&(ab)->map_inaddr, ina, sizeof(struct in_addr));		\
	memset((ab)->map_zero, 0, sizeof((ab)->map_zero));		\
	memset((ab)->map_one, 0xff, sizeof((ab)->map_one));		\
}

union inx_addr {
	struct in_addr	in_addr;
	struct in6_addr	in6_addr;
	struct {
		u_char	mau_zero[10];
		u_char	mau_one[2];
		struct in_addr mau_inaddr;
	}		map_addr_un;
#define	map_zero	map_addr_un.mau_zero
#define	map_one		map_addr_un.mau_one
#define	map_inaddr	map_addr_un.mau_inaddr
};

static struct hostent *_hpcopy(struct hostent *hp);
#ifdef MAPPED_ADDR_ENABLED
static struct hostent *_hpmerge(struct hostent *hp1, struct hostent *hp2);
#endif /* MAPPED_ADDR_ENABLED */

static struct hostent *files_ghbyname(const char *name, int af);
static struct hostent *files_ghbyaddr(const void *addr, int addrlen, int af);
#ifdef DNS
static struct hostent *dns_ghbyname(const char *name, int af);
static struct hostent *dns_ghbyaddr(const void *addr, int addrlen, int af);
#endif /* DNS */
#ifdef ICMPNL
static struct hostent *icmp_ghbyaddr(const void *addr, int addrlen, int af);
#endif /* ICMPNL */

/*
 * Functions defined in draft-ietf-ipngwg-bsd-api-new-01.txt
 *	getnodebyname, getnodebyaddr, freehostent
 */

static struct hostent *
_ghbyname(const char *name, int af, int flags)
{
	struct hostent *hp;

	if (flags & AI_ADDRCONFIG) {
		int s;

		if ((s = socket(af, SOCK_DGRAM, 0)) < 0) {
			h_errno = NO_RECOVERY;
			return NULL;
		}
		/*
		 * TODO:
		 * Note that implementation dependent test for address
		 * configuration should be done everytime called,
		 * because addresses will be dynamically assigned or deleted.
		 */
		close(s);
	}

	if ((hp = files_ghbyname(name, af)) != NULL)
		return hp;
#ifdef DNS
	if ((hp = dns_ghbyname(name, af)) != NULL)
		return hp;
#endif /* DNS */

	return NULL;
}

struct hostent *
getnodebyname(const char *name, int af, int flags)
{
	struct hostent *hp, hpbuf;
	char *addrs[2];
	union inx_addr addrbuf;
#ifdef MAPPED_ADDR_ENABLED
	int laf;
	struct hostent *hp2;
#endif /* MAPPED_ADDR_ENABLED */

	if (af != AF_INET && af != AF_INET6) {
		h_errno = NO_RECOVERY;
		return NULL;
	}

	/* special case for literal address */
#ifdef MAPPED_ADDR_ENABLED
	if (inet_pton((laf = AF_INET6), name, &addrbuf) > 0
	||  inet_pton((laf = AF_INET), name, &addrbuf) > 0) {
		if (af == AF_INET && laf == AF_INET6) {
			h_errno = HOST_NOT_FOUND;
			return NULL;
		}
		if (af == AF_INET6 && laf == AF_INET) {
			MAPADDR(&addrbuf, &addrbuf.in_addr);
			h_errno = HOST_NOT_FOUND;
			return NULL;
		}
#else /* MAPPED_ADDR_ENABLED */ /*}*/
	if (inet_pton(af, name, &addrbuf) > 0) {
#endif /* MAPPED_ADDR_ENABLED */
		hp = &hpbuf;
		hp->h_name = (char *)name;
		hp->h_aliases = NULL;
		hp->h_addrtype = af;
		hp->h_length = ADDRLEN(af);
		hp->h_addr_list = addrs;
		addrs[0] = (char *)&addrbuf;
		addrs[1] = NULL;
		return _hpcopy(hp);
	}

	h_errno = HOST_NOT_FOUND;
	hp = _ghbyname(name, af, flags);

#ifdef MAPPED_ADDR_ENABLED
	if (af == AF_INET6
	&&  ((flags & AI_ALL) || ((flags & AI_V4MAPPED) && hp == NULL))) {
		if ((hp2 = _ghbyname(name, AF_INET, flags)) != NULL) {
			if (hp == NULL) {
				/* dummy to convert IPv6 mapped addr */
				memset(&hpbuf, 0, sizeof(hpbuf));
				hpbuf.h_addrtype = af;
				hpbuf.h_length = ADDRLEN(af);
				if ((hp = _hpcopy(&hpbuf)) == NULL) {
					freehostent(hp2);
					return NULL;
				}
			}
			hp = _hpmerge(hp, hp2);
		}
	}
#endif /* MAPPED_ADDR_ENABLED */
	return hp;
}

struct hostent *
getnodebyaddr(const void *src, size_t len, int af)
{
	struct hostent *hp;

	if (af != AF_INET && af != AF_INET6) {
		h_errno = NO_RECOVERY;
		return NULL;
	}

	if (af == AF_INET6 && len == sizeof(struct in6_addr)
	&&  (IN6_IS_ADDR_V4MAPPED((struct in6_addr *)src))) {
		src = (char *)src +
			(sizeof(struct in6_addr) - sizeof(struct in_addr));
		af = AF_INET;
		len = sizeof(struct in_addr);
	}
	h_errno = HOST_NOT_FOUND;

	if ((hp = files_ghbyaddr(src, len, af)) != NULL)
		return hp;
#ifdef ICMPNL
	if ((hp = icmp_ghbyaddr(src, len, af)) != NULL)
		return hp;
#endif /* ICMPNL */
#ifdef DNS
	if ((hp = dns_ghbyaddr(src, len, af)) != NULL)
		return hp;
#endif /* DNS */

	return NULL;
}

void
freehostent(struct hostent *ptr)
{
	free(ptr);
}

/*
 * Functions for compatibility.
 *	gethostbyname, gethostbyname2, gethostbyaddr
 */

/* hostent pointer for non thread-safe functions */
static struct hostent *saved_hp;

struct hostent *
gethostbyname2(const char *name, int af)
{
	struct hostent *hp, hpbuf;
	char *addrs[2];
	union inx_addr addrbuf;

	if (saved_hp != NULL)
		freehostent(saved_hp);

	/* special case for literal address */
	/* XXX: use inet_aton to handle nonstandard format for IPv4 */
	if ((af == AF_INET ? inet_aton(name, &addrbuf.in_addr) :
			     inet_pton(af, name, &addrbuf)) > 0) {
		hp = &hpbuf;
		hp->h_name = (char *)name;
		hp->h_aliases = NULL;
		hp->h_addrtype = af;
		hp->h_length = ADDRLEN(af);
		hp->h_addr_list = addrs;
		addrs[0] = (char *)&addrbuf;
		addrs[1] = NULL;
		return saved_hp = _hpcopy(hp);
	}

	h_errno = HOST_NOT_FOUND;
	return saved_hp = _ghbyname(name, af, 0);
}

struct hostent *
gethostbyname(const char *name)
{

	return gethostbyname2(name, AF_INET);
}

struct hostent *
gethostbyaddr(const char *src, int len, int af)
{

	if (saved_hp != NULL)
		freehostent(saved_hp);
	return saved_hp = getnodebyaddr(src, len, af);
}

/*
 * Private utility functions
 */

/*
 * _hpcopy: allocate and copy hostent structure
 */
static struct hostent *
_hpcopy(struct hostent *hp)
{
	struct hostent *nhp;
	char *cp, **pp;
	int size, addrsize;
	int nalias = 0, naddr = 0;
	int al_off;
	int i;

	if (hp == NULL) {
		h_errno = HOST_NOT_FOUND;
		return hp;
	}

	/* count size to be allocated */
	size = sizeof(struct hostent);
	if (hp->h_name != NULL && *hp->h_name != '\0')
		size += strlen(hp->h_name) + 1;
	if ((pp = hp->h_aliases) != NULL) {
		for (i = 0; *pp != NULL; i++, pp++) {
			if (**pp != '\0') {
				size += strlen(*pp) + 1;
				nalias++;
			}
		}
	}
	/* adjust alignment */
	size = ALIGN(size);
	al_off = size;
	if (nalias != 0)
		size += sizeof(char *) * (nalias + 1);
	addrsize = ALIGN(hp->h_length);
	if ((pp = hp->h_addr_list) != NULL) {
		while (*pp++ != NULL)
			naddr++;
		if (naddr != 0) {
			size += addrsize * naddr;
			size += sizeof(char *) * (naddr + 1);
		}
	}

	/* copy */
	if ((nhp = (struct hostent *)malloc(size)) == NULL) {
		h_errno = TRY_AGAIN;
		return NULL;
	}
	cp = (char *)&nhp[1];
	if (hp->h_name != NULL && *hp->h_name != '\0') {
		nhp->h_name = cp;
		strcpy(cp, hp->h_name);
		cp += strlen(cp) + 1;
	} else
		nhp->h_name = NULL;
	if (nalias != 0) {
		nhp->h_aliases = (char **)((char *)nhp + al_off);
		for (i = 0, pp = hp->h_aliases; *pp != NULL; pp++) {
			if (**pp != '\0') {
				nhp->h_aliases[i++] = cp;
				strcpy(cp, *pp);
				cp += strlen(cp) + 1;
			}
		}
		nhp->h_aliases[i++] = NULL;
		cp = (char *)&nhp->h_aliases[i];
	} else {
		nhp->h_aliases = NULL;
		cp = (char *)nhp + al_off;
	}
	nhp->h_addrtype = hp->h_addrtype;
	nhp->h_length = hp->h_length;
	if (naddr != 0) {
		nhp->h_addr_list = (char **)cp;
		cp += sizeof(char *) * (naddr + 1);
		for (i = 0, pp = hp->h_addr_list; *pp != NULL; pp++) {
			nhp->h_addr_list[i++] = cp;
			memcpy(cp, *pp, hp->h_length);
			cp += addrsize;
		}
		nhp->h_addr_list[i] = NULL;
	} else
		nhp->h_addr_list = NULL;
	return nhp;
}

#ifdef MAPPED_ADDR_ENABLED
/*
 * _hpmerge: merge 2 hostent structure, arguments will be freed
 */
static struct hostent *
_hpmerge(struct hostent *hp1, struct hostent *hp2)
{
	int i, j;
	int naddr, nalias;
	char **pp;
	struct hostent *hp, hpbuf;
	char *aliases[MAXALIASES + 1], *addrs[MAXADDRS + 1];
	union inx_addr addrbuf[MAXADDRS];

	if (hp1 == NULL)
		return hp2;
	if (hp2 == NULL)
		return hp1;

#define	HP(i)	(i == 1 ? hp1 : hp2)
	hp = &hpbuf;
	hp->h_name = (hp1->h_name != NULL ? hp1->h_name : hp2->h_name);
	hp->h_aliases = aliases;
	nalias = 0;
	for (i = 1; i <= 2; i++) {
		if ((pp = HP(i)->h_aliases) == NULL)
			continue;
		while (nalias < MAXALIASES && *pp != NULL) {
			/* check duplicates */
			for (j = 0; j < nalias; j++)
				if (strcasecmp(*pp, aliases[j]) == 0)
					break;
			if (j == nalias)
				aliases[nalias++] = *pp++;
		}
	}
	aliases[nalias] = NULL;
	if (hp1->h_length != hp2->h_length) {
		hp->h_addrtype = AF_INET6;
		hp->h_length = sizeof(struct in6_addr);
	} else {
		hp->h_addrtype = hp1->h_addrtype;
		hp->h_length = hp1->h_length;
	}
	hp->h_addr_list = addrs;
	naddr = 0;
	for (i = 1; i <= 2; i++) {
		if ((pp = HP(i)->h_addr_list) == NULL)
			continue;
		if (HP(i)->h_length == hp->h_length) {
			while (naddr < MAXADDRS && *pp != NULL)
				addrs[naddr++] = *pp++;
		} else {
			/* copy IPv4 addr as mapped IPv6 addr */
			while (naddr < MAXADDRS && *pp != NULL) {
				MAPADDR(&addrbuf[naddr], *pp++);
				addrs[naddr] = (char *)&addrbuf[naddr];
				naddr++;
			}
		}
	}
	addrs[naddr] = NULL;
	hp = _hpcopy(hp);
	freehostent(hp1);
	freehostent(hp2);
	return hp;
}
#endif /* MAPPED_ADDR_ENABLED */

/*
 * FILES (/etc/hosts)
 */
static char *
_getword(char **pp)
{
	char c, *p, *ret;
	const char *sp;
	static const char sep[] = " \t\n";

	ret = NULL;
	for (p = *pp; (c = *p) != '\0'; p++) {
		for (sp = sep; *sp != '\0'; sp++) {
			if (c == *sp)
				break;
		}
		if (ret == NULL) {
			if (*sp == '\0')
				ret = p;
		} else {
			if (*sp != '\0') {
				*p++ = '\0';
				break;
			}
		}
	}
	*pp = p;
	if (ret == NULL || *ret == '\0')
		return NULL;
	return ret;
}

static struct hostent *
files_ghbyname(const char *name, int af)
{
	int match, nalias;
	char *p, *line, *addrstr, *cname;
	FILE *fp;
	struct hostent *hp, hpbuf;
	char *aliases[MAXALIASES + 1], *addrs[2];
	union inx_addr addrbuf;
	char buf[BUFSIZ];

#ifdef __V6D__
	if ((p = getenv("V6ROOT")) != NULL) {
		strcpy(buf, p);
		strcat(buf, _PATH_HOSTS);
		fp = fopen(buf, "r");
	} else
#endif /* __V6D__ */
	fp = fopen(_PATH_HOSTS, "r");
	if (fp == NULL) {
		h_errno = NO_RECOVERY;
		return NULL;
	}
	hp = NULL;

	while (fgets(buf, sizeof(buf), fp)) {
		line = buf;
		if (*line == '#')
			continue;
		if ((p = strchr(line, '#')) != NULL)
			*p = '\0';
		if ((addrstr = _getword(&line)) == NULL
		||  (cname = _getword(&line)) == NULL)
			continue;
		match = (strcasecmp(cname, name) == 0);
		nalias = 0;
		while ((p = _getword(&line)) != NULL) {
			if (!match)
				match = (strcasecmp(p, name) == 0);
			if (nalias < MAXALIASES)
				aliases[nalias++] = p;
		}
		if (!match)
			continue;
		if (inet_pton(af, addrstr, &addrbuf) <= 0) {
			h_errno = NO_DATA;	/* name found */
			continue;
		}
		hp = &hpbuf;
		hp->h_name = cname;
		hp->h_aliases = aliases;
		aliases[nalias] = NULL;
		hp->h_addrtype = af;
		hp->h_length = ADDRLEN(af);
		hp->h_addr_list = addrs;
		addrs[0] = (char *)&addrbuf;
		addrs[1] = NULL;
		hp = _hpcopy(hp);
		break;
	}
	fclose(fp);
	return hp;
}

static struct hostent *
files_ghbyaddr(const void *addr, int addrlen, int af)
{
	int nalias;
	char *p, *line;
	FILE *fp;
	struct hostent *hp, hpbuf;
	char *aliases[MAXALIASES + 1], *addrs[2];
	union inx_addr addrbuf;
	char buf[BUFSIZ];

#ifdef __V6D__
	if ((p = getenv("V6ROOT")) != NULL) {
		strcpy(buf, p);
		strcat(buf, _PATH_HOSTS);
		fp = fopen(buf, "r");
	} else
#endif /* __V6D__ */
	fp = fopen(_PATH_HOSTS, "r");
	if (fp == NULL) {
		h_errno = NO_RECOVERY;
		return NULL;
	}
	hp = NULL;
	while (fgets(buf, sizeof(buf), fp)) {
		line = buf;
		if (*line == '#')
			continue;
		if ((p = strchr(line, '#')) != NULL)
			*p = '\0';
		if ((p = _getword(&line)) == NULL
		||  inet_pton(af, p, &addrbuf) <= 0
		||  memcmp(addr, &addrbuf, addrlen) != 0
		||  (p = _getword(&line)) == NULL)
			continue;
		hp = &hpbuf;
		hp->h_name = p;
		hp->h_aliases = aliases;
		nalias = 0;
		while ((p = _getword(&line)) != NULL) {
			if (nalias < MAXALIASES)
				aliases[nalias++] = p;
		}
		aliases[nalias] = NULL;
		hp->h_addrtype = af;
		hp->h_length = addrlen;
		hp->h_addr_list = addrs;
		addrs[0] = (char *)&addrbuf;
		addrs[1] = NULL;
		hp = _hpcopy(hp);
		break;
	}
	fclose(fp);
	return hp;
}

#ifdef DNS

#include <arpa/nameser.h>
#include <resolv.h>

#ifdef DEBUG
#define	ASSERT(X)	if (!(X)) fprintf(stderr, "ASSFAIL: %s %d: %s\n", __FILE__, __LINE__, #X);
#else
#define	ASSERT(X)
#endif

static struct hostent *
dns_ghbyname(const char *name, int af)
{
	int n;
	u_char answer[BUFSIZ];
	char tbuf[MAXDNAME+1];
	HEADER *hp;
	u_char *cp, *eom;
	int qtype;
	int type, class, ancount, qdcount;
	u_long ttl;
	char hostbuf[BUFSIZ];
	char *bp;
	char *alist[MAXALIASES];
	char *hlist[MAXADDRS];
	struct hostent hbuf;
	int buflen;
	int na, nh;

	if ((_res.options & RES_INIT) == 0) {
		if (res_init() < 0)
			return NULL;
	}
	hbuf.h_aliases = NULL;
	hbuf.h_addrtype = af;
	hbuf.h_length = ADDRLEN(af);
	hbuf.h_addr_list = hlist;
	na = nh = 0;

	qtype = (af == AF_INET6 ? T_AAAA : T_A);
	n = res_search(name, C_IN, qtype, answer, sizeof(answer));
	if (n < 0) {
		return NULL;
	}
	hp = (HEADER *)answer;
	eom = answer + n;
	ancount = ntohs(hp->ancount);
	qdcount = ntohs(hp->qdcount);
	ASSERT(qdcount == 1);
	cp = answer + sizeof(HEADER);
	bp = hostbuf;
	buflen = sizeof(hostbuf);

	n = dn_expand(answer, eom, cp, bp, buflen);
	ASSERT(n >= 0);
	cp += n + QFIXEDSZ;
	hbuf.h_name = bp;
	n = strlen(bp) + 1;
	bp += n;
	buflen -= n;
	while (ancount-- > 0 && cp < eom) {
		n = dn_expand(answer, eom, cp, bp, buflen);
		ASSERT(n >= 0);
		cp += n;		/* name */
		type = _getshort(cp);
		cp += 2;		/* type */
		class = _getshort(cp);
		cp += 2;		/* class */
		ttl = _getlong(cp);
		cp += 4;		/* ttl */
		n = _getshort(cp);
		cp += 2;		/* len */
		ASSERT(class == C_IN);
		switch (type) {
		case T_CNAME:
			if (na >= MAXALIASES-1) {
				cp += n;
				break;
			}
			n = dn_expand(answer, eom, cp, tbuf, sizeof(tbuf));
			ASSERT(n >= 0);
			cp += n;
			/* alias */
			if (na == 0)
				hbuf.h_aliases = alist;
			alist[na++] = bp;
			n = strlen(bp) + 1;
			bp += n;
			buflen -= n;
			/* canon */
			n = strlen(tbuf) + 1;
			ASSERT(n < buflen);
			strcpy(bp, tbuf);
			hbuf.h_name = bp;
			bp += n;
			buflen -= n;
			break;
		case T_A:
		case T_AAAA:
if (type != qtype) { printf("Got other type!\n"); cp += n; break; }
			bp = (char *)ALIGN(bp);
			ASSERT(n == hbuf.h_length);
			ASSERT(n < buflen);
			if (nh < MAXADDRS-1) {
				hlist[nh++] = bp;
				memcpy(bp, cp, n);
				bp += n;
				buflen -= n;
			}
			cp += n;
			break;
		default:
			ASSERT(0);
			cp += n;
			break;
		}
	}
	if (nh == 0) {
		h_errno = NO_RECOVERY;
		return NULL;
	}
	alist[na] = NULL;
	hlist[nh] = NULL;
	return _hpcopy(&hbuf);
}

static struct hostent *
dns_ghbyaddr(const void *addr, int addrlen, int af)
{
	int n;
	u_char answer[BUFSIZ];
	HEADER *hp;
	u_char c, *cp, *eom;
	int type, class, ancount, qdcount;
	u_long ttl;
	char hostbuf[BUFSIZ];
	char *bp;
	char *alist[MAXALIASES];
	char *hlist[2];
	struct hostent hbuf;
	int buflen;
	int na;
	static char hex[] = "0123456789abcdef";

	/* XXX */
	if (af == AF_INET6 && IN6_IS_ADDR_LINKLOCAL((struct in6_addr *)addr))
		return NULL;

	if ((_res.options & RES_INIT) == 0) {
		if (res_init() < 0)
			return NULL;
	}
	hbuf.h_name = NULL;
	hbuf.h_aliases = NULL;
	hbuf.h_addrtype = af;
	hbuf.h_length = addrlen;
	hbuf.h_addr_list = hlist;
	hlist[0] = (char *)addr;
	hlist[1] = NULL;
	na = 0;

	if (af == AF_INET6) {
		for (n = 0, cp = (u_char *)addr+sizeof(struct in6_addr)-1,
					bp = hostbuf;
				n < sizeof(struct in6_addr);
				n++, cp--, bp += 4) {
			bp[0] = hex[*cp & 0xf];
			bp[1] = '.';
			bp[2] = hex[*cp >> 4];
			bp[3] = '.';
		}
		strcpy(bp, "ip6.int");
	} else {
		for (n = 0, cp = (u_char *)addr+sizeof(struct in_addr)-1,
					bp = hostbuf;
				n < sizeof(struct in_addr);
				n++, cp--) {
			c = *cp;
			if (c >= 100)
				*bp++ = '0' + c / 100;
			if (c >= 10)
				*bp++ = '0' + (c % 100) / 10;
			*bp++ = '0' + c % 10;
			*bp++ = '.';
		}
		strcpy(bp, "in-addr.arpa");
	}

	n = res_search(hostbuf, C_IN, T_PTR, answer, sizeof(answer));
	if (n < 0) {
		return NULL;
	}
	hp = (HEADER *)answer;
	eom = answer + n;
	ancount = ntohs(hp->ancount);
	qdcount = ntohs(hp->qdcount);
	ASSERT(qdcount == 1);
	cp = answer + sizeof(HEADER);
	bp = hostbuf;
	buflen = sizeof(hostbuf);

	n = dn_expand(answer, eom, cp, bp, buflen);
	ASSERT(n >= 0);
	cp += n + QFIXEDSZ;
	while (ancount-- > 0 && cp < eom) {
		n = dn_expand(answer, eom, cp, bp, buflen);
		ASSERT(n >= 0);
		cp += n;		/* name */
		type = _getshort(cp);
		cp += 2;		/* type */
		class = _getshort(cp);
		cp += 2;		/* class */
		ttl = _getlong(cp);
		cp += 4;		/* ttl */
		n = _getshort(cp);
		cp += 2;		/* len */
		ASSERT(class == C_IN);
		switch (type) {
		case T_PTR:
			n = dn_expand(answer, eom, cp, bp, buflen);
			ASSERT(n >= 0);
			cp += n;
			if (na >= MAXALIASES-1)
				break;
			if (hbuf.h_name == NULL) {
				hbuf.h_name = bp;
			} else {
				if (na == 0)
					hbuf.h_aliases = alist;
				alist[na++] = bp;
			}
			n = strlen(bp) + 1;
			bp += n;
			buflen -= n;
			break;
		default:
			ASSERT(0);
			/*FALLTHRU*/
		case T_CNAME:
			cp += n;
			break;
		}
	}
	if (hbuf.h_name == NULL)
		return NULL;
	alist[na] = NULL;
	return _hpcopy(&hbuf);
}
#endif /* DNS */

#ifdef ICMPNL

/*
 * experimental:
 *	draft-ietf-ipngwg-icmp-namelookups-02.txt
 *	ifindex is assumed to be encoded in addr.
 */
#include <sys/uio.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>

static char *
icmp_fqdn_query(const struct in6_addr *addr, int ifindex)
{
	int s;
	struct icmp6_filter filter;
	struct msghdr msg;
	struct cmsghdr *cmsg;
	struct in6_pktinfo *pkt;
	char cbuf[256];
	char buf[1024];
	int cc;
	struct icmp6_fqdn_query *fq;
	struct icmp6_fqdn_reply *fr;
	static struct host_cache {
		struct host_cache *hc_next;
		int hc_ifindex;
		struct in6_addr hc_addr;
		char *hc_name;
	} *hc_head = NULL;
	struct host_cache *hc;
	static int pid;
	struct sockaddr_in6 sin6;
	struct iovec iov;
	fd_set s_fds, fds;
	struct timeval tout;
	int len;
	char *name;

	for (hc = hc_head; hc; hc = hc->hc_next) {
		if (hc->hc_ifindex == ifindex
		&&  IN6_ARE_ADDR_EQUAL(&hc->hc_addr, addr))
			return hc->hc_name;
	}

	hc = (struct host_cache *)malloc(sizeof(*hc));
	if (hc == NULL)
		return NULL;
	hc->hc_ifindex = ifindex;
	hc->hc_addr = *addr;
	hc->hc_name = NULL;
	hc->hc_next = hc_head;
	hc_head = hc;

	if (pid == 0)
		pid = getpid();

	ICMP6_FILTER_SETBLOCKALL(&filter);
	ICMP6_FILTER_SETPASS(ICMP6_FQDN_REPLY, &filter);

	FD_ZERO(&s_fds);
	tout.tv_sec = 0;
	tout.tv_usec = 200000;	/*XXX: 200ms*/

	fq = (struct icmp6_fqdn_query *)buf;
	fq->icmp6_fqdn_type = ICMP6_FQDN_QUERY;
	fq->icmp6_fqdn_code = 0;
	fq->icmp6_fqdn_cksum = 0;
	fq->icmp6_fqdn_id = (u_short)pid;
	fq->icmp6_fqdn_unused = 0;
	fq->icmp6_fqdn_cookie[0] = 0;
	fq->icmp6_fqdn_cookie[1] = 0;

	memset(&sin6, 0, sizeof(sin6));
	sin6.sin6_family = AF_INET6;
	sin6.sin6_addr = *addr;

	memset(&msg, 0, sizeof(msg));
	msg.msg_name = (caddr_t)&sin6;
	msg.msg_namelen = sizeof(sin6);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	iov.iov_base = (caddr_t)buf;
	iov.iov_len = sizeof(struct icmp6_fqdn_query);

	if (ifindex) {
		msg.msg_control = cbuf;
		msg.msg_controllen = sizeof(cbuf);
		cmsg = CMSG_FIRSTHDR(&msg);
		cmsg->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
		cmsg->cmsg_level = IPPROTO_IPV6;
		cmsg->cmsg_type = IPV6_PKTINFO;
		pkt = (struct in6_pktinfo *)&cmsg[1];
		memset(&pkt->ipi6_addr, 0, sizeof(struct in6_addr));
		pkt->ipi6_ifindex = ifindex;
		cmsg = CMSG_NXTHDR(&msg, cmsg);
		msg.msg_controllen = (char *)cmsg - cbuf;
	}

	s = socket(PF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
	if (s < 0)
		return NULL;
	(void)setsockopt(s, IPPROTO_ICMPV6, ICMP6_FILTER,
			 (char *)&filter, sizeof(filter));
	cc = sendmsg(s, &msg, 0);
	if (cc < 0) {
		close(s);
		return NULL;
	}
	FD_SET(s, &s_fds);
	for (;;) {
		fds = s_fds;
		if (select(s + 1, &fds, NULL, NULL, &tout) <= 0) {
			close(s);
			return NULL;
		}
		len = sizeof(sin6);
		cc = recvfrom(s, buf, sizeof(buf), 0,
			      (struct sockaddr *)&sin6, &len);
		if (cc <= 0) {
			close(s);
			return NULL;
		}
		if (cc < sizeof(struct ip6_hdr) + sizeof(struct icmp6_hdr))
			continue;
		if (!IN6_ARE_ADDR_EQUAL(addr, &sin6.sin6_addr))
			continue;
		fr = (struct icmp6_fqdn_reply *)(buf + sizeof(struct ip6_hdr));
		if (fr->icmp6_fqdn_type == ICMP6_FQDN_REPLY)
			break;
	}
	close(s);
	if (fr->icmp6_fqdn_cookie[1] != 0) {
		/* rfc1788 type */
		name = buf + sizeof(struct ip6_hdr) + sizeof(struct icmp6_hdr) + 4;
		len = (buf + cc) - name;
	} else {
		len = fr->icmp6_fqdn_namelen;
		name = fr->icmp6_fqdn_name;
	}
	if (len <= 0)
		return NULL;
	name[len] = 0;

	hc->hc_name = strdup(name);
	return hc->hc_name;
}

static struct hostent *
icmp_ghbyaddr(const void *addr, int addrlen, int af)
{
	struct hostent hbuf, *hp;
	char *hname, *alist[2];
	int ifindex;
	struct in6_addr addr6;

	if (af != AF_INET6) {
		/*
		 * Note: rfc1788 defines Who Are You for IPv4,
		 * but no one implements it.
		 */
		return NULL;
	}

	memcpy(&addr6, addr, addrlen);
	ifindex = addr6.s6_addr[3];
	addr6.s6_addr[3] = 0;

	if (!IN6_IS_ADDR_LINKLOCAL(&addr6))
		return NULL;	/*XXX*/

	hname = icmp_fqdn_query(&addr6, ifindex);
	if (hname == NULL)
		return NULL;

	hp = &hbuf;
	hp->h_name = hname;
	hp->h_aliases = NULL;
	hp->h_addrtype = af;
	hp->h_length = addrlen;
	hp->h_addr_list = alist;
	alist[0] = (char *)addr;
	alist[1] = NULL;
	return _hpcopy(hp);
}
#endif /* ICMPNL */
