/*
 * Copyright (C) 1997 and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 *		PPP Multiprotocol common procedures
 *
 * $Id: multiproto.c,v 1.1.4.1 1998/04/20 13:03:23 shin Exp $
 *
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#ifdef INET6
#include <netinet/ip6.h>
#endif /* INET6 */
#include <netinet/in_var.h>
#include <netinet/ip_icmp.h>
#ifdef INET6
#include <netinet/icmp6.h>
#endif /* INET6 */
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#ifndef NOALIAS
#include <alias.h>
#endif /* NOALIAS */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "defs.h"
#include "timer.h"
#include "fsm.h"
#include "command.h"
#include "hdlc.h"
#include "lcp.h"
#include "multiproto.h"
#ifdef INET6
#include "ipv6.h"
#endif /* INET6 */
#include "loadalias.h"
#include "vars.h"
#include "id.h"
#include "filter.h"
#include "mbuf.h"
#include "systems.h"
#include "log.h"
#include "os.h"
#include "route.h"

struct afsw afsw[PPP_AFMAX];
int af_sock2ppp[AF_MAX];
int af_ppp2sock[PPP_AFMAX];
    
char ntop_buf[
#ifdef INET6
	      INET6_ADDRSTRLEN
#elif INET
	      INET_ADDRSTRLEN
#endif /* INET6, INET */
	      ];
int reconnectState, reconnectCount;
int AfActiveFlag;

static int KeepAlive;
static struct pppTimer IdleTimer;

static int ProtoLinkIsUp(void);
static void ProtoLinkdown(u_long af);

void 
IdleTimeout(void *v)
{
  LogPrintf(LogPHASE, "Idle timer expired.\n");
  reconnect(RECON_FALSE);
  LcpClose();
}

/*
 *  Start Idle timer. If timeout is reached, we call LcpClose() to
 *  close LCP and link.
 */
void
StartIdleTimer(void)
{
  static time_t IdleStarted;
	
  if (!(mode & (MODE_DEDICATED | MODE_DDIAL))) {
    StopTimer(&IdleTimer);
    IdleTimer.func = IdleTimeout;
    IdleTimer.load = VarIdleTimeout * SECTICKS;
    IdleTimer.state = TIMER_STOPPED;
    time(&IdleStarted);
    IdleTimer.arg = (void *)&IdleStarted;    
    StartTimer(&IdleTimer);
  }
}

void
UpdateIdleTimer(void)
{
  if (ProtoLinkIsUp())
    StartIdleTimer();
}

void
StopIdleTimer(void)
{
  StopTimer(&IdleTimer);
}

int
RemainingIdleTime()
{
  if (VarIdleTimeout == 0 || IdleTimer.state != TIMER_RUNNING ||
      IdleTimer.arg == NULL)
    return -1;
  return VarIdleTimeout - (time(NULL) - *(time_t *)IdleTimer.arg);
}

/*
 *  If any protocol layer traffic is detected, refresh IdleTimer.
 */
void
RestartIdleTimer(u_long af)
{
  if (!(mode & (MODE_DEDICATED | MODE_DDIAL)) && KeepAlive) {
    time((time_t *)IdleTimer.arg);	  
    StartTimer(&IdleTimer);
  }
}

static u_short interactive_ports[32] = {
  544, 513, 514, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 21, 22, 23, 0, 0, 0, 0, 0, 0, 0, 543,
};

#define	INTERACTIVE(p)	(interactive_ports[(p) & 0x1F] == (p))

static char *TcpFlags[] = {
  "FIN", "SYN", "RST", "PSH", "ACK", "URG",
};

static char *Direction[] = {"INP", "OUT", "OUT", "IN/OUT"};
static struct filterent *Filters[] = {ifilters, ofilters, dfilters, afilters};

static int
PortMatch(int op, u_short pport, u_short rport)
{
  switch (op) {
    case OP_EQ:
    return (pport == rport);
  case OP_GT:
    return (pport > rport);
  case OP_LT:
    return (pport < rport);
  default:
    return (0);
  }
}

extern void IpInit(void);
extern void Ipv6Init(void);

/*
 * Init af_flags[]
 */
void
AfInit(void)
{
#ifdef INET
  IpInit();
#endif /* INET */
#ifdef INET6
  Ipv6Init();
#endif /* INET6 */
}

/*
 *  Check a packet against with defined filters
 */
static int
FilterCheck(u_long af, void *vp, int direction)
{
  struct filterent *fp = Filters[direction];
  int gotinfo, cproto, estab, n;
  struct tcphdr *th;
  struct udphdr *uh;
  struct icmp *ih;
  char *ptop;
  u_short sport, dport;
  struct ip *pip;
  
  switch (af) {
#ifdef INET
  case AF_INET:
    pip = (struct ip *) vp;
    break;
#endif /* INET */
#ifdef INET6
  case AF_INET6:
    /* XXX; TODO: support filtering for IPv6 */
    return (A_PERMIT);
    break;
#endif /* INET6 */  
  default:
    return (A_DENY);
    /* NOTREACHED */
    break;
  }    

  if (fp->action) {
    cproto = gotinfo = estab = 0;
    sport = dport = 0;
    for (n = 0; n < MAXFILTERS; n++) {
      if (fp->action) {
	/* permit fragments on in and out filter */
	if ((direction == FL_IN || direction == FL_OUT) &&
	    (ntohs(pip->ip_off) & IP_OFFMASK) != 0) {
	  return (A_PERMIT);
	}
	LogPrintf(LogDEBUG, "rule = %d\n", n);
	if ((pip->ip_src.s_addr & fp->smask.s_addr) ==
	    (fp->saddr.s_addr & fp->smask.s_addr) &&
	    (pip->ip_dst.s_addr & fp->dmask.s_addr) ==
	    (fp->daddr.s_addr & fp->dmask.s_addr)) {
	  if (fp->proto) {
	    if (!gotinfo) {
	      ptop = (char *) pip + (pip->ip_hl << 2);

	      switch (pip->ip_p) {
	      case IPPROTO_ICMP:
		cproto = P_ICMP;
		ih = (struct icmp *) ptop;
		sport = ih->icmp_type;
		estab = 1;
		break;
	      case IPPROTO_UDP:
		cproto = P_UDP;
		uh = (struct udphdr *) ptop;
		sport = ntohs(uh->uh_sport);
		dport = ntohs(uh->uh_dport);
		estab = 1;
		break;
	      case IPPROTO_TCP:
		cproto = P_TCP;
		th = (struct tcphdr *) ptop;
		sport = ntohs(th->th_sport);
		dport = ntohs(th->th_dport);
		estab = (th->th_flags & TH_ACK);
		if (estab == 0)
		  LogPrintf(LogDEBUG, "flag = %02x, sport = %d, dport = %d\n",
			    th->th_flags, sport, dport);
		break;
	      default:
		return (A_DENY);/* We'll block unknown type of packet */
	      }
	      gotinfo = 1;
	      LogPrintf(LogDEBUG, "dir = %d, proto = %d, srcop = %d,"
			" dstop = %d, estab = %d\n", direction, cproto,
			fp->opt.srcop, fp->opt.dstop, estab);
	    }
	    LogPrintf(LogDEBUG, "check0: rule = %d, proto = %d, sport = %d,"
		      " dport = %d\n", n, cproto, sport, dport);
	    LogPrintf(LogDEBUG, "check0: action = %d\n", fp->action);

	    if (cproto == fp->proto) {
	      if ((fp->opt.srcop == OP_NONE ||
		   PortMatch(fp->opt.srcop, sport, fp->opt.srcport))
		  &&
		  (fp->opt.dstop == OP_NONE ||
		   PortMatch(fp->opt.dstop, dport, fp->opt.dstport))
		  &&
		  (fp->opt.estab == 0 || estab)) {
		return (fp->action);
	      }
	    }
	  } else {
	    /* Address is mached. Make a decision. */
	    LogPrintf(LogDEBUG, "check1: action = %d\n", fp->action);
	    return (fp->action);
	  }
	}
      }
      fp++;
    }
    return (A_DENY);		/* No rule is mached. Deny this packet */
  }
  return (A_PERMIT);		/* No rule is given. Permit this packet */
}


/* 
 * Logging sub routine for PacketCheck
 */
static int
LogPacketType(char *logbuf, int loglen, char *pr_name, u_long af, void *src, void *dst, 
	      u_short type_or_sport, u_short type_or_dport)
{
  snprintf(logbuf + loglen, sizeof logbuf - loglen,
	   "%s: %s:%d ---> ", pr_name, inet_ntop(af, src, ntop_buf, sizeof(ntop_buf)), type_or_sport);
  loglen += strlen(logbuf + loglen);
  snprintf(logbuf + loglen, sizeof logbuf - loglen,
	   "%s:%d", inet_ntop(af, dst, ntop_buf, sizeof(ntop_buf)), type_or_dport);
  loglen += strlen(logbuf + loglen);
  return (loglen);
}
/*
 *  For debugging aid.
 */
int
PacketCheck(u_long af, void *vp, int nb, int direction)
{

  char *ptop;
#if defined(INET) || defined(INET6)
  struct tcphdr *th;
  struct udphdr *uh;
#endif /* INET || INET6 */
#ifdef INET
  struct ip *pip;
  struct icmp *icmph;
#endif /* INET */
#ifdef INET6
  struct ip6_hdr *pip6;
  struct icmp6_hdr *icmp6h;
  int inet6_flag = 0; /* to enable checking of extension headers */
#endif /* INET6 */
  void *src, *dst;

  int mask, len, n;
  int pri = PRI_NORMAL;
  int logit, loglen;
  static char logbuf[200];
  u_char nh;

  logit = LogIsKept(LogTCPIP) && direction != FL_DIAL;
  loglen = 0;

  switch (af) {
#ifdef INET
  case AF_INET:
    pip = (struct ip *) vp;
    ptop = (vp + (pip->ip_hl << 2));
    nh = pip->ip_p;
    src = &pip->ip_src;
    dst = &pip->ip_dst;
    break;
#endif /* INET */
#ifdef INET6
  case AF_INET6:
    pip6 = (struct ip6_hdr *)vp;
    ptop = ((char *)vp + sizeof(struct ip6_hdr));
    nh = pip6->ip6_nxt;
    src = &pip6->ip6_src;
    dst = &pip6->ip6_dst;
    inet6_flag = 1;
    break;
#endif /* INET6 */
  default:
    return (-1);
    /* NOTREACHED */
    break;
  }

  if (logit && loglen < sizeof logbuf) {
    snprintf(logbuf + loglen, sizeof logbuf - loglen, "%s ",
	     Direction[direction]);
    loglen += strlen(logbuf + loglen);
  }

  switch (nh) {

#ifdef INET
  case IPPROTO_ICMP:
    if (logit && loglen < sizeof logbuf) {
      icmph = (struct icmp *) ptop;
      loglen = LogPacketType(logbuf, loglen, "ICMP", af, src, dst,
			     icmph->icmp_type, icmph->icmp_type);
    }
    break;
#endif /* INET */

#ifdef INET6
  case IPPROTO_ICMPV6:
    if (logit && loglen < sizeof logbuf) {
      icmp6h = (struct icmp6_hdr *) ptop;
      loglen = LogPacketType(logbuf, loglen, "ICMP6", af, src, dst,
			     icmp6h->icmp6_type, icmp6h->icmp6_type);
    }
    break;
#endif /* INET6 */

#if defined(INET) || defined(INET6)
  case IPPROTO_UDP:
    if (logit && loglen < sizeof logbuf) {
      uh = (struct udphdr *) ptop;
      loglen = LogPacketType(logbuf, loglen, "UDP", af, src, dst,
			     ntohs(uh->uh_sport), ntohs(uh->uh_dport));
    }
    break;

  case IPPROTO_TCP:
    th = (struct tcphdr *) ptop;
#ifdef INET
    if (af == AF_INET) {
      if (((struct ip *)vp)->ip_tos == IPTOS_LOWDELAY)
	pri = PRI_FAST;
      else if ((ntohs(((struct ip *)vp)->ip_off) & IP_OFFMASK) == 0) {
	if (INTERACTIVE(ntohs(th->th_sport)) ||
	    INTERACTIVE(ntohs(th->th_dport)))
	  pri = PRI_FAST;
      }
    }
#endif /* INET */
    if (logit && loglen < sizeof logbuf) {
      /* init len */
      switch (af) {
#ifdef INET
      case AF_INET:
	len = ntohs(((struct ip *)vp)->ip_len) - 
	  (((struct ip *)vp)->ip_hl << 2) - (th->th_off << 2);
	break;
#endif /* INET */
#ifdef INET6
      case AF_INET6:
	len = ntohs(((struct ip6_hdr *)vp)->ip6_plen) - (th->th_off << 2);
	break;
#endif /* INET6 */
      default:
	/* should not reach here */
	len = -1;
	break;
      }
      loglen = LogPacketType(logbuf, loglen, "TCP", af, src, dst, ntohs(th->th_sport),
			     ntohs(th->th_dport));
      n = 0;
      for (mask = TH_FIN; mask != 0x40; mask <<= 1) {
	if (th->th_flags & mask) {
	  snprintf(logbuf + loglen, sizeof logbuf - loglen, " %s", TcpFlags[n]);
	  loglen += strlen(logbuf + loglen);
	}
	n++;
      }
      snprintf(logbuf + loglen, sizeof logbuf - loglen,
	       "  seq:%x  ack:%x (%d/%d)",
	       ntohl(th->th_seq), ntohl(th->th_ack), len, nb);
      loglen += strlen(logbuf + loglen);
      if ((th->th_flags & TH_SYN) && nb > 40) {
	u_short *sp;

	ptop += 20;
	sp = (u_short *) ptop;
	if (ntohs(sp[0]) == 0x0204) {
	  snprintf(logbuf + loglen, sizeof logbuf - loglen,
		   " MSS = %d", ntohs(sp[1]));
	  loglen += strlen(logbuf + loglen);
	}
      }
    }
    break;
#endif /* INET || INET6 */

  default:
    if (logit && loglen < sizeof logbuf) {
      loglen = LogPacketType(logbuf, loglen, "unknown(IPv6 ext hdr ??):", af, src, dst, 0, 0);
    }
    break;
  }

  if (logit)
    LogPrintf(LogTCPIP, "%s\n", logbuf);

  if ((FilterCheck(af, vp, direction) & A_DENY)) {
    LogPrintf(LogDEBUG, "blocked.\n");
    if (direction == 0 && afsw[af_sock2ppp[af]].AfErrorReport)
      (*afsw[af_sock2ppp[af]].AfErrorReport)(vp, pri);
    return (-1);
  } else {
    if (FilterCheck(af, vp, FL_KEEP) & A_DENY) {
      if (logit)
	LogPrintf(LogTCPIP, "%s - NO KEEPALIVE\n", logbuf);  
      KeepAlive = FALSE;
    } else {
      if (logit)
	LogPrintf(LogTCPIP, "%s\n", logbuf);  	    
      KeepAlive = TRUE;
    }
	    
    return (pri);
  }
}

void
AfEnqueue(struct afsw *afsw_p, int pri, char *ptr, int count)
{
  struct mbuf *bp;

  bp = mballoc(count, afsw_p->AfMbQType);
  memcpy(MBUF_CTOP(bp), ptr, count);
  Enqueue(&afsw_p->AfOutputQueues[pri], bp);
}

static void
ProtoRemoveAddr(void)
{
  int i;

  for (i = PPP_AFMIN; i < PPP_AFMAX; i++)
    if (afsw[i].AfResetAddress)
      (*afsw[i].AfResetAddress)();
}

/* 
 * Do protocol dependent initialization after the first protocol dependent link is up.
 */
void
ProtoLinkup(u_long af)
{
  if ((afsw[af_sock2ppp[af]].AfFlags & PPP_AFFLAGS_LINKUP) == 0) {
    const char *s;
    int Level;
      
    reconnectState = RECON_UNKNOWN;
    if (mode & MODE_BACKGROUND && BGFiledes[1] != -1) {
      char c = EX_NORMAL;

      if (write(BGFiledes[1], &c, 1) == 1)
	LogPrintf(LogPHASE, "Parent notified of success.\n");
      else
	LogPrintf(LogPHASE, "Failed to notify parent of success.\n");
      close(BGFiledes[1]);
      BGFiledes[1] = -1;
    }

    s = inet_ntop(af, afsw[af_sock2ppp[af]].AfHisAddr, ntop_buf,
		  sizeof(ntop_buf));
    Level = LogIsKept(LogLINK) ? LogLINK : afsw[af_sock2ppp[af]].AfLogLevel;
    LogPrintf(Level, "ProtoLinkup: %s\n", s);

    if (SelectSystem(inet_ntop(af, afsw[af_sock2ppp[af]].AfMyAddr,
			       ntop_buf, sizeof(ntop_buf)),
		     LINKUPFILE) < 0) {
      if (!GetLabel() || SelectSystem(GetLabel(), LINKUPFILE) < 0) {
#ifdef INET
        /* for backward compatibility */
	if (af == AF_INET)
	  SelectSystem("MYADDR", LINKUPFILE);
#endif /* INET */
        SelectSystem(afsw[af_sock2ppp[af]].AfMyAddrLabel, LINKUPFILE);
      }
    }
    afsw[af_sock2ppp[af]].AfFlags |= PPP_AFFLAGS_LINKUP;
  }
}

/* 
 * Check if any of protocol dependent link is up.
 */
static int
ProtoLinkIsUp(void)
{
  int i;

  for (i = PPP_AFMIN; i < PPP_AFMAX; i++)
    if (afsw[i].AfFlags & PPP_AFFLAGS_LINKUP)
      return (1);

  return 0;
}

/* 
 * Close all of protocol dependent link, and CCP link.
 */
void
PppLinkdown(void)
{
  int i, linkup = 0;

  for (i = PPP_AFMIN; i < PPP_AFMAX; i++)
    if (afsw[i].AfFlags & PPP_AFFLAGS_LINKUP) {
      linkup = 1;
      ProtoLinkdown(af_ppp2sock[i]);
    }

  if (linkup)
    FsmDown(&CcpFsm);	/* CCP must come down */
}
	
/* Close one of protocol dependent links. */
static void
ProtoLinkdown(u_long af)
{
  char *s;
  int Level;

  s = (char *) inet_ntop(af, afsw[af_sock2ppp[af]].AfHisAddr, ntop_buf,
			 sizeof(ntop_buf));
  Level = LogIsKept(LogLINK) ? LogLINK : afsw[af_sock2ppp[af]].AfLogLevel;
  LogPrintf(Level, "ProtoLinkdown: %s\n", s);
  
  FsmDown(afsw[af_sock2ppp[af]].AfFsm);
  afsw[af_sock2ppp[af]].AfFlags &= ~PPP_AFFLAGS_LINKUP;
  if (SelectSystem(s, LINKDOWNFILE) < 0) {
    if (!GetLabel() || SelectSystem(GetLabel(), LINKDOWNFILE) < 0) {
#ifdef INET
      /* for backward compatibility */
      if (af == AF_INET && SelectSystem("MYADDR", LINKDOWNFILE) < 0)
#endif /* INET */
	SelectSystem(afsw[af_sock2ppp[af]].AfMyAddrLabel, LINKDOWNFILE);
    }
  }
}

extern int tunno;

/*
 * Call PppLinkdown() and after that, remove all of protocol dependent addresses.
 */
int
ProtoInterfaceDown(int final)
{
  int s;

  PppLinkdown();
  if (!final && (mode & MODE_DAEMON))	/* We still want interface alive */
    return (0);
  s = socket(DEF_SOCK_AF, SOCK_DGRAM, 0);
  if (s < 0) {
    LogPrintf(LogERROR, "ProtoInterfaceDown: socket: %s\n", strerror(errno));
    return (-1);
  }
  ifrq.ifr_flags &= ~IFF_UP;
  if (ID0ioctl(s, SIOCSIFFLAGS, &ifrq) < 0) {
    LogPrintf(LogERROR, "ProtoInterfaceDown: ioctl(SIOCSIFFLAGS): %s\n",
	      strerror(errno));
    close(s);
    return (-1);
  }
  ProtoRemoveAddr();

  close(s);
  return (0);
}

void
ProtoInitLenFromSockMask(u_long af, struct sockaddr *sa)
{
  int i, socklen;
  char *mp;

  mp = ((char *)sa) + afsw[af_sock2ppp[af]].AfAddrOffInSock;

  for (i = 0,socklen = 0; *mp; i++,mp++) {
    socklen++;
  }
  sa->sa_len = socklen ? socklen + afsw[af_sock2ppp[af]].AfAddrOffInSock : 0;
}

/* 
 * Send each protocols packet in que.
 */
void
ProtoStartOutput(void)
{
  int i;

  for (i = PPP_AFMIN; i < PPP_AFMAX; i++)
    if (afsw[i].AfStartOutput) {
      (*afsw[i].AfStartOutput)();
    }
}
